from dataclasses import dataclass

from ... import helper
from ...model.type.xpath import XPath


@dataclass(kw_only=True, slots=True)
class LoginForm1Step:
    """Settings for login form page in 1 step where both username and password are displayed at once.

    url: Optional URL to login page.

    post_login_url_contains: Optionally await redirect to this URL as a user is typically redirected automatically to a new page or view after logging in.

    post_login_element_xpath: Upon successful login, optionally await this element to be loaded."""

    username_input_xpath: str
    password_input_xpath: str
    submit_button_xpath: str  # Specific for this class.

    # Shared defaults:
    url: str | None = None
    post_login_wait_seconds: int | None = None
    post_login_url_contains: str | None = None
    post_login_element_xpath: str | None = None

    def __post_init__(self) -> None:
        self.username_input_xpath = XPath(self.username_input_xpath)
        self.password_input_xpath = XPath(self.password_input_xpath)
        self.submit_button_xpath = XPath(self.submit_button_xpath)
        self.url = helper.url.mediate_conversion_to_tiny_type_or_none(self.url)
        self.post_login_element_xpath = helper.xpath.mediate_conversion_to_tiny_type_or_none(
            self.post_login_element_xpath)


@dataclass(kw_only=True, slots=True)
class LoginForm2Steps:
    """Settings for login form page in 2 steps where username is prompted first, and once confirmed, then the password can be entered.

    url: Optional URL to login page.

    post_login_url_contains: Optionally await redirect to this URL as a user is typically redirected automatically to a new page or view after logging in.

    post_login_element_xpath: Upon successful login, optionally await this element to be loaded."""

    username_input_xpath: str
    username_submit_button_xpath: str  # Specific for this class.
    password_input_xpath: str
    password_submit_button_xpath: str  # Specific for this class.

    # Shared defaults:
    url: str | None = None
    post_login_wait_seconds: int | None = None
    post_login_url_contains: str | None = None
    post_login_element_xpath: str | None = None

    def __post_init__(self) -> None:
        self.username_input_xpath = XPath(self.username_input_xpath)
        self.username_submit_button_xpath = XPath(self.username_submit_button_xpath)
        self.password_input_xpath = XPath(self.password_input_xpath)
        self.password_submit_button_xpath = XPath(self.password_submit_button_xpath)
        self.url = helper.url.mediate_conversion_to_tiny_type_or_none(self.url)
        self.post_login_element_xpath = helper.xpath.mediate_conversion_to_tiny_type_or_none(
            self.post_login_element_xpath)



from ...exception.headless import MethodNotSupportedInHeadlessModeException
from ...model.browser.base.driver import BrowserDriver
from ...model.browser.base.settings import BrowserSettings
from ...model.driver_methods import DriverMethods
from .input_field import select_input_field


class SelectDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def input_field(self, xpath: str) -> None:
        """Select input field, similar to clicking the mouse on a form field."""

        if self._settings.headless:
            raise MethodNotSupportedInHeadlessModeException(
                "select.input_field", "headless mode doesn't support interactions")

        select_input_field(self._driver, xpath)



from ..browser.get.page_title import get_page_title
from ..browser.get.url.current import get_current_url


class WaitForElementTimeoutException(Exception):
    __slots__ = ["message"]

    def __init__(self, driver: object, xpath: str) -> None:
        current_url = get_current_url(driver)
        self.message = f"On page {current_url}, waiting for element timed out: {xpath}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class WaitForPageTitleToChangeTimeoutException(Exception):
    __slots__ = ["message"]

    def __init__(self, driver: object, page_title_or_fragment: str) -> None:
        current_url = get_current_url(driver)
        page_title = get_page_title(driver)
        self.message = f"On page {current_url} with title \"{page_title}\", check of page title fragment timed out, or the title fragment doesn't match: {page_title_or_fragment}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class WaitForUrlTimeoutException(Exception):
    __slots__ = ["message"]

    def __init__(self, driver: object, url: str) -> None:
        current_url = get_current_url(driver)
        self.message = f"On page {current_url}, waiting for URL timed out, or the URL fragment doesn't exist: {url}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class WaitForWindowTimeoutException(Exception):
    __slots__ = ["message"]

    def __init__(self) -> None:
        self.message = "Waiting for new tab or window to open timed out."
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message



from abc import ABC, abstractmethod

from ..base.driver import BrowserDriver


class BrowserExtension(ABC):
    """Abstract class that contains the extension methods for certain browser types."""

    __slots__ = ["browser_driver"]

    def __init__(self, browser_driver: BrowserDriver) -> None:
        self.browser_driver = browser_driver

    @abstractmethod
    def enable_images(self) -> None:
        """Intended use: Revert disabled images for a browser when its web driver doesn't support a session-based configuration for Selenium, but only a global configuration that impacts normal use."""



from ..... import helper
from .....constant import timeout
from .....model.type.url import URL
from ....get.url.current import get_current_url


def wait_until_url_changes(driver: object, baseline_url: str, timeout: int = timeout.DEFAULT) -> None:
    def has_url_changed(driver: object, baseline_url: str) -> bool:
        return get_current_url(driver) != baseline_url

    baseline_url = URL(baseline_url)
    helper.retry.until_condition_is_true(driver, baseline_url, func=has_url_changed, timeout=timeout)



import re

from ..... import constant, helper
from .....constant import timeout
from .....model.type.xpath import XPath
from ....get.text import get_text
from ...for_element import wait_for_element


def wait_until_text_contains(driver: object, xpath: str, regex: str, timeout: int = timeout.DEFAULT) -> None:
    def does_element_text_contain(driver: object, xpath: str, regex: str) -> bool:
        text = get_text(driver, xpath, constant.timeout.BYPASS)
        return bool(re.search(regex, text, re.IGNORECASE))

    xpath = XPath(xpath)
    wait_for_element(driver, xpath, timeout)
    helper.retry.until_condition_is_true(driver, xpath, regex, func=does_element_text_contain, timeout=timeout)



from ....constant import timeout
from ....model.browser.base.driver import BrowserDriver
from ....model.browser.base.settings import BrowserSettings
from ....model.driver_methods import DriverMethods
from .element_disappears import wait_until_element_disappears
from .images_have_loaded import wait_until_images_have_loaded
from .number_of_window_handles_is import wait_until_number_of_window_handles_is
from .page_title.__main__ import WaitUntilPageTitleDriverMethods
from .text.__main__ import WaitUntilTextDriverMethods
from .url.__main__ import WaitUntilUrlDriverMethods


class WaitUntilDriverMethods(DriverMethods):
    __slots__ = ["page_title", "text", "url"]

    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)
        self.page_title: WaitUntilPageTitleDriverMethods = WaitUntilPageTitleDriverMethods(browser_driver, settings)
        self.text: WaitUntilTextDriverMethods = WaitUntilTextDriverMethods(browser_driver, settings)
        self.url: WaitUntilUrlDriverMethods = WaitUntilUrlDriverMethods(browser_driver, settings)

    def element_disappears(self, xpath: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until element doesn't exist."""

        wait_until_element_disappears(self._driver, xpath, timeout)

    def images_have_loaded(self, xpath: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until element doesn't exist. The image XPath can target one or more images."""

        wait_until_images_have_loaded(self._driver, xpath, timeout)

    def number_of_window_handles_is(self, timeout: int = timeout.DEFAULT) -> None:
        """Wait until number of window handles is."""

        wait_until_number_of_window_handles_is(self._driver, timeout)



from ...model.browser.base.driver import BrowserDriver
from ...model.browser.base.settings import BrowserSettings
from ...model.driver_methods import DriverMethods
from .switch_to import switch_to_iframe
from .switch_to_original_page import switch_to_original_page


class IframeDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def switch_to(self, xpath: str) -> None:
        """Switch to iframe."""

        switch_to_iframe(self._driver, xpath)

    def switch_to_original_page(self) -> None:
        """After switch to iframe, use this to go back to the original page."""

        switch_to_original_page(self._driver)



from typing import Any

import lxml.etree  # type: ignore
from lxml.etree import XPathSyntaxError

from ..model.type.xpath import XPath


def is_valid(xpath: str) -> bool:
    try:
        return bool(lxml.etree.XPath(xpath))
    except (XPathSyntaxError, Exception):
        return False


def set_attributes(self: Any, name: str, value: Any, attributes: list[str]) -> Any:
    """Intended for the __setattr__ dunder method in data classes where some string arguments need to be validated and set as XPath.

    Note that this is only supported for data classes when __dict__ is used and not __slots__."""

    if name and value:
        self.__dict__[name] = XPath(value) if name in attributes else value
    return self


def mediate_conversion_to_tiny_type_or_none(value: str | None) -> XPath | None:
    """Mediate conversion of string to XPath tiny type or keep None type."""

    return None if value is None else XPath(value)



from __future__ import annotations

from ... import helper
from ...exception.url import URLSyntaxError


class URL(str):
    """Class to handle and validate URL input as "tiny type"."""

    __slots__ = ["value"]

    def __new__(cls, url: str | URL) -> URL:
        # If input already is a validated URL element, bypass and don't create a new object:
        return url if isinstance(url, URL) else super().__new__(cls, url)

    def __init__(self, url: str) -> None:
        if not helper.url.is_valid(url):
            raise URLSyntaxError(url)
        self.value: str = url

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return self.value

    def is_valid(self) -> bool:
        return helper.url.is_valid(self.value)



def get_total_scroll_height(driver: object) -> int:
    script_get_total_scroll_height = """return Math.max(
        document.body.scrollHeight, document.documentElement.scrollHeight,
        document.body.offsetHeight, document.documentElement.offsetHeight,
        document.body.clientHeight, document.documentElement.clientHeight
    );"""
    return int(driver.execute_script(script_get_total_scroll_height))  # type: ignore



from ....model.browser.base.driver import BrowserDriver
from ....model.browser.base.settings import BrowserSettings
from ....model.driver_methods import DriverMethods
from .position import get_window_position
from .size import get_window_size


class WindowGetDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def position(self) -> tuple[int, int]:
        """Get the coordinates of the top left corner of the browser window on the screen. Usage:

        x, y = browser.window.get.position()"""

        return get_window_position(self._driver)

    def size(self) -> tuple[int, int]:
        """Get width and height of browser window on the screen in pixels. Usage:

        width, height = browser.window.get.size()"""

        return get_window_size(self._driver)



from dataclasses import dataclass

from ....constant import directory
from ....helper import operating_system
from .page_load_strategy import PageLoadStrategy
from .type import BrowserType


@dataclass(kw_only=True, slots=True)
class BrowserSettings:
    """Class to configure the browser driver.

    headless: If enabled, note that some interactable methods, e.g. "select", aren't supported.

    screenshot_dir: Destination directory for screenshot files. If not set, default directory is from where the script is executed."""

    type: BrowserType = BrowserType.EDGE if operating_system.is_windows() else BrowserType.CHROME
    headless: bool = False
    disable_images: bool = False
    page_load_strategy: PageLoadStrategy = PageLoadStrategy.NORMAL
    path_to_executable: str | None = None
    screenshot_dir: str = directory.CURRENT



import time

from ...constant import timeout


def scroll_to_position(driver: object, x: int, y: int) -> None:
    driver.execute_script(f"window.scrollTo({x}, {y});")  # type: ignore
    time.sleep(timeout.VERY_SHORT)  # Small delay to ensure the view is updated.



from selenium.webdriver.common.by import By

from ...constant import timeout
from ...model.type.xpath import XPath
from ..wait.for_element import wait_for_element


def select_input_field(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> None:
    xpath = XPath(xpath)
    wait_for_element(driver, xpath, timeout)
    driver.find_element(By.XPATH, xpath).click()  # type: ignore



import time

from ....constant import timeout
from ...get.screen.height import get_screen_height
from ..by import scroll_by


def scroll_page_up(driver: object) -> None:
    screen_height = get_screen_height(driver)
    y_scroll_pixels = -(screen_height + 1)
    scroll_by(driver, 0, y_scroll_pixels)
    time.sleep(timeout.VERY_SHORT)  # Small delay to ensure the view is updated.



from urllib.parse import urlparse

from ..model.type.url import URL


def ensure_trailing_slash(url: str) -> str:
    """When comparing URLs, e.g. "http://example.com/" and "http://example.com", use this method to normalise the comparison."""

    if "?" in url:  # If the URL contains a parameter (e.g. https://example.com/search?page=1), ignore trailing slash.
        return url
    return url if url[-1] == "/" else f"{url}/"


def is_https(url: str) -> bool:
    return url.startswith("https:")


def is_valid(url: str) -> bool:
    if url.startswith("file://"):  # Accept files from local machine as valid URL and as an exception.
        return True
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except Exception:
        return False


def mediate_conversion_to_tiny_type_or_none(value: str | None) -> URL | None:
    """Mediate conversion of string to URL tiny type or keep None type."""

    return None if value is None else URL(value)


def mediate_https(url1: str, url2: str) -> tuple[str, str]:
    """If a URL is HTTPS, the other URL should also be HTTPS."""

    if (is_https(url1) and is_https(url2)) or (not is_https(url1) and not is_https(url2)):
        return url1, url2
    elif not is_https(url1):
        url1 = url1.replace("http:", "https:")
    elif not is_https(url2):
        url2 = url2.replace("http:", "https:")
    return url1, url2


def remove_parameters(url: str) -> str:
    return url if "?" not in url else url.split("?")[0]



import time

from ...constant import timeout
from ...model.combo_settings.cookie_banner import CookieBannerSettings
from ..click.button import click_button
from ..open.url_if_not_current import open_url_if_not_current
from ..wait.for_element import wait_for_element
from ..wait.until.element_disappears import wait_until_element_disappears


def combo_cookie_banner(driver: object, settings: CookieBannerSettings) -> None:
    if settings.url is not None:
        open_url_if_not_current(driver, settings.url)
    if settings.has_loaded_wait_seconds is not None:
        time.sleep(settings.has_loaded_wait_seconds)
    if settings.has_loaded_xpath is not None:
        wait_for_element(driver, settings.has_loaded_xpath)
    click_button(driver, settings.button_xpath)
    if settings.has_disappeared_wait_seconds is not None:
        time.sleep(settings.has_disappeared_wait_seconds)
    else:
        time.sleep(timeout.VERY_SHORT)
    wait_until_element_disappears(driver, settings.button_xpath)



from ...model.browser.base.driver import BrowserDriver
from ...model.browser.base.settings import BrowserSettings
from ...model.driver_methods import DriverMethods
from ...model.window.controller import WindowHandleController
from .close import window_close
from .fullscreen import window_fullscreen
from .get.__main__ import WindowGetDriverMethods
from .handle.__main__ import WindowHandleDriverMethods
from .maximize import window_maximize
from .minimize import window_minimize
from .open.__main__ import WindowOpenDriverMethods
from .set.__main__ import WindowSetDriverMethods
from .switch_to import switch_to_window


class WindowDriverMethods(DriverMethods):
    __slots__ = ["_controller", "get", "handle", "open", "set"]

    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)
        self._controller: WindowHandleController = WindowHandleController(self._driver)
        self.get: WindowGetDriverMethods = WindowGetDriverMethods(browser_driver, settings)
        self.handle: WindowHandleDriverMethods = WindowHandleDriverMethods(browser_driver, settings, self._controller)
        self.open: WindowOpenDriverMethods = WindowOpenDriverMethods(browser_driver, settings, self._controller)
        self.set: WindowSetDriverMethods = WindowSetDriverMethods(browser_driver, settings)

    def close(self) -> None:
        """Close current tab or window."""

        window_close(self._driver, self._controller)

    def fullscreen(self) -> None:
        """Fills the entire screen, similar to pressing F11 in most browsers."""

        window_fullscreen(self._driver)

    def maximize(self) -> None:
        """Enlarges the window. For most operating systems, the window will fill the screen, without blocking the operating system's own menus and toolbars."""

        window_maximize(self._driver)

    def minimize(self) -> None:
        """Minimizes the window of current browsing context. The exact behavior of this command is specific to individual window managers. Minimize Window typically hides the window in the system tray."""

        window_minimize(self._driver)

    def switch_to(self, window_handle: str) -> None:
        """Switch to window/tab by handle ID or name."""

        switch_to_window(self._driver, self._controller, window_handle)

    def switch_to_original_window(self) -> None:
        """Switch to initial window/tab."""

        original_window_handle_id = self._controller.get_handle_id_by_name(self._controller._original_window_name)
        switch_to_window(self._driver, self._controller, original_window_handle_id)



from selenium.common.exceptions import TimeoutException
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait  # type: ignore

from ....constant import timeout
from ....exception.timeout import WaitForWindowTimeoutException


def wait_until_number_of_window_handles_is(driver: object, expected_handles: int, timeout: int = timeout.DEFAULT) -> None:
    try:
        WebDriverWait(driver, timeout).until(EC.number_of_windows_to_be(expected_handles))  # type: ignore
    except TimeoutException:
        raise WaitForWindowTimeoutException() from TimeoutException



from ....model.browser.base.driver import BrowserDriver
from ....model.browser.base.settings import BrowserSettings
from ....model.driver_methods import DriverMethods
from .position import set_window_position
from .size import set_window_size


class WindowSetDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def position(self, x: int, y: int) -> None:
        """Moves the window to the chosen coordinate of the screen in pixels."""

        set_window_position(self._driver, x, y)

    def size(self, width: int, height: int) -> None:
        """Restores the window and sets the window size."""

        set_window_size(self._driver, width, height)



from __future__ import annotations

from ... import helper
from ...exception.xpath import XPathSyntaxError


class XPath(str):
    """Class to handle and validate XPath input as "tiny type"."""

    __slots__ = ["value"]

    def __new__(cls, xpath: str | XPath) -> XPath:
        # If input already is a validated XPath element, bypass and don't create a new object:
        return xpath if isinstance(xpath, XPath) else super().__new__(cls, xpath)

    def __init__(self, xpath: str) -> None:
        if not helper.xpath.is_valid(xpath):
            raise XPathSyntaxError(xpath)
        self.value: str = xpath

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return self.value

    def is_valid(self) -> bool:
        return helper.xpath.is_valid(self.value)



from ....model.browser.base.driver import BrowserDriver
from ....model.browser.base.settings import BrowserSettings
from ....model.driver_methods import DriverMethods
from .down import scroll_page_down
from .to_end import scroll_to_end_of_page
from .to_top import scroll_to_top_of_page
from .up import scroll_page_up


class ScrollPageDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def down(self) -> None:
        """If possible, scroll page down."""

        scroll_page_down(self._driver)

    def to_end(self) -> None:
        """If possible, scroll to end of page."""

        scroll_to_end_of_page(self._driver)

    def to_top(self) -> None:
        """If possible, scroll to top of page."""

        scroll_to_top_of_page(self._driver)

    def up(self) -> None:
        """If possible, scroll page up."""

        scroll_page_up(self._driver)



import re


def tool_is_input_valid(text: str, regex: str, ignore_case: bool = True) -> bool:
    if ignore_case:
        match = re.fullmatch(regex, text, re.IGNORECASE)
    else:
        match = re.fullmatch(regex, text)
    return bool(match)



__all__ = ["directory", "image", "operating_system", "retry", "screenshot", "url", "window_handle", "xpath"]

from . import directory, image, operating_system, retry, screenshot, url, window_handle, xpath



from ....model.browser.base.driver import BrowserDriver
from ....model.browser.base.settings import BrowserSettings
from ....model.driver_methods import DriverMethods
from .height import get_screen_height
from .size import get_screen_size
from .width import get_screen_width


class GetScreenSizeDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def height(self) -> int:
        """Get inner height of the screen in pixels."""

        return get_screen_height(self._driver)

    def size(self) -> tuple[int, int]:
        """Get inner width and height of the screen in pixels. Usage:

        width, height = browser.get.screen_size()"""

        return get_screen_size(self._driver)

    def width(self) -> int:
        """Get inner width of the screen in pixels."""

        return get_screen_width(self._driver)



from .....constant import timeout
from .....model.browser.base.driver import BrowserDriver
from .....model.browser.base.settings import BrowserSettings
from .....model.driver_methods import DriverMethods
from .changes import wait_until_page_title_changes
from .contains import wait_until_page_title_contains
from .equals import wait_until_page_title_equals


class WaitUntilPageTitleDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def changes(self, baseline_text: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until the page title changes from a baseline text, e.g. after a page reload or change. The text is evaluated as an exact match."""

        wait_until_page_title_changes(self._driver, baseline_text, timeout)

    def contains(self, page_title_fragment: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until the page title has changed, e.g. after a redirect or update. The input can contain both a fragment or the full page title."""

        wait_until_page_title_contains(self._driver, page_title_fragment, timeout)

    def equals(self, page_title: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until the page title has changed, e.g. after a redirect or update. The input has to match the exact page title."""

        wait_until_page_title_equals(self._driver, page_title, timeout)



from selenium.webdriver.common.by import By

from ...model.type.xpath import XPath
from ..wait.for_element import wait_for_element


def switch_to_iframe(driver: object, xpath: str) -> None:
    xpath = XPath(xpath)
    wait_for_element(driver, xpath)
    iframe_element = driver.find_element(By.XPATH, xpath)  # type: ignore
    driver.switch_to.frame(iframe_element)  # type: ignore



from .... import factory
from .__main__ import BrowserExtension


class SafariBrowserExtension(BrowserExtension):
    def enable_images(self) -> None:
        """Intended use: Revert disabled images for Safari as its web driver doesn't support a session-based configuration for Selenium, but only a global configuration that impacts normal use."""

        factory.safari.enable_images(self.browser_driver)



from ..by import scroll_by
from ..get.position import get_scroll_position

Y_TEST_DISTANCE = 1


def check_if_scroll_is_end_of_page(driver: object) -> bool:
    # Let's assume that we can't scroll further than the end of the page...
    _, y_initial = get_scroll_position(driver)
    # ... and so scrolling down should yield the same position as the initial one.
    scroll_by(driver, 0, Y_TEST_DISTANCE)
    _, y_after = get_scroll_position(driver)
    # Remember to reset the scroll position to initial position:
    scroll_by(driver, 0, -Y_TEST_DISTANCE)
    return y_initial == y_after



from ....model.window.controller import WindowHandleController


def get_all_window_handles(driver: object, controller: WindowHandleController, selenium: bool = False) -> list[str]:
    if selenium:
        return driver.window_handles  # type: ignore
    else:
        return [window_handle.id for window_handle in controller._window_handles]



from ... import helper
from ...model.type.url import URL
from ..get.url.current import get_current_url
from .url import open_url


def open_url_if_not_current(driver: object,
                            url: str,
                            ignore_trailing_slash: bool = True,
                            ignore_parameters: bool = False,
                            ignore_https: bool = False) -> None:
    url = URL(url)
    current_url = get_current_url(driver)
    if ignore_https:
        current_url, url = helper.url.mediate_https(current_url, url)
    if ignore_parameters:
        current_url = helper.url.remove_parameters(current_url)
        url = helper.url.remove_parameters(url)
    if ignore_trailing_slash:
        current_url = helper.url.ensure_trailing_slash(current_url)
        url = helper.url.ensure_trailing_slash(url)
    if current_url != url:
        open_url(driver, url)



BYPASS: int = 0
"""No timeout."""

VERY_SHORT: int = 1
"""Very short timeout: 1 second."""

SHORT: int = 3
"""Short timeout: 3 seconds."""

DEFAULT: int = 5
"""Default timeout: 5 seconds."""

LONG: int = 10
"""Long timeout: 10 seconds."""



from ....model.browser.base.driver import BrowserDriver
from ....model.browser.base.settings import BrowserSettings
from ....model.driver_methods import DriverMethods
from ....model.window.controller import WindowHandleController
from ....model.window.tab_or_window import TabOrWindow
from .new_tab_or_window import open_new_tab_or_window


class WindowOpenDriverMethods(DriverMethods):
    __slots__ = ["_controller"]

    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings, controller: WindowHandleController) -> None:
        super().__init__(browser_driver, settings)
        self._controller = controller

    def new_tab(self, url: str | None = None, name: str | None = None) -> None:
        """Open and switch to new tab in current window. The URL and name of the tab are optional arguments.

        name: Can be used to switch to this tab with the "browser.window.switch_to(name)" method."""

        open_new_tab_or_window(self._driver, self._controller, TabOrWindow.TAB, url, name)

    def new_window(self, url: str | None = None, name: str | None = None) -> None:
        """Open and switch to new window. The URL and name of the window are optional arguments.

        name: Can be used to switch to this window with the "browser.window.switch_to(name)" method."""

        open_new_tab_or_window(self._driver, self._controller, TabOrWindow.WINDOW, url, name)



class WindowHandleIdNotFoundError(Exception):
    __slots__ = ["message"]

    def __init__(self, id: str) -> None:
        self.message = f"Window handle ID not found or doesn't exist: {id}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class WindowHandleNameNotFoundError(Exception):
    __slots__ = ["message"]

    def __init__(self, name: str) -> None:
        self.message = f"Window handle name not found or doesn't exist: {name}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class WindowHandleIdNotUniqueError(Exception):
    __slots__ = ["message"]

    def __init__(self, id: str) -> None:
        self.message = f"Window handle ID already exists: {id}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class WindowHandleNameNotUniqueError(Exception):
    __slots__ = ["message"]

    def __init__(self, name: str) -> None:
        self.message = f"Window handle name already exists: {name}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class WindowHandleIdNotValidError(Exception):
    __slots__ = ["message"]

    def __init__(self, id: str) -> None:
        self.message = f"Window handle ID has invalid format: {id}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class WindowHandleNameNotValidError(Exception):
    __slots__ = ["message"]

    def __init__(self, name: str) -> None:
        self.message = f"Window handle name is invalid. Try using another name than this: {name}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message



from ....model.browser.base.driver import BrowserDriver
from ....model.browser.base.settings import BrowserSettings
from ....model.driver_methods import DriverMethods
from .visible_portion import get_screenshot_visible_portion


class GetScreenshotDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)

    def visible_portion(self, file_name: str | None = None, destination_dir: str | None = None) -> None:
        """Take screenshot of visible portion and save as PNG image. Default destination directory is from where the script is executed. Examples:

        browser.get.screenshot.visible_portion() # Default file name and destination

        browser.get.screenshot.visible_portion("image.png") # Custom file name and default destination

        browser.get.screenshot.visible_portion("image.png", "./screenshots") # Custom file name and destination

        browser.get.screenshot.visible_portion(destination_dir = "./screenshots") # Default file name and custom destination"""

        get_screenshot_visible_portion(self._driver, self._settings, file_name, destination_dir)



import re

from ..model.window.handle import WindowHandle


def name_already_exists(name: str, window_handles: list[WindowHandle]) -> bool:
    return name in [window_handle.name for window_handle in window_handles]


def id_already_exists(id: str, window_handles: list[WindowHandle]) -> bool:
    return id in [window_handle.id for window_handle in window_handles]


WINDOW_HANDLE_ID_PATTERN = re.compile(r"^CDwindow-[A-Z0-9]{32}$")


def is_valid_id(id: str) -> bool:
    """Checks if handle ID is similar to this pattern: CDwindow-8088CB616D7499360039D98453AE91FC"""

    return bool(WINDOW_HANDLE_ID_PATTERN.match(id))



import os
import re
import urllib

from ..constant import directory
from . import operating_system


def create_if_not_exists(dir_name: str) -> None:
    if dir_name == directory.CURRENT:
        return
    if not os.path.exists(dir_name):
        os.mkdir(dir_name)


def ensure_trailing_slash(dir_name: str) -> str:
    return dir_name if dir_name[-1] == "/" else f"{dir_name}/"


def encode_path_as_url(path: str) -> str:
    url_encoded_path = str(urllib.parse.quote(path, safe="/:"))
    return url_encoded_path.replace("%2520", "%20")  # Handle edge cases.


def ensure_windows_file_path_format_encoding_as_url(path: str) -> str:
    """Relevant for Windows where a file path name should be "file://path/to/file.html" instead of "file://path\\to\\file.html" ."""

    output = path.replace("\\", "/")  # Raw replace backslash with slash.
    # Handle exception for "file:///C:/path/to/file.html" declaration in URLs:
    if re.match(r"^file:/+[A-Za-z]:", output, re.IGNORECASE):
        output = re.sub(r"^file:/+", "file:///", output, re.IGNORECASE)
    return encode_path_as_url(output)


def update_path_format_if_windows(path: str) -> str:
    return ensure_windows_file_path_format_encoding_as_url(path) if operating_system.is_windows() else encode_path_as_url(path)



from ....model.window.controller import WindowHandleController
from .all import get_all_window_handles


def count_window_handles(driver: object, controller: WindowHandleController, selenium: bool = False) -> int:
    if selenium:
        return len(get_all_window_handles(driver, controller, selenium=True))
    else:
        return controller.count()



from ...constant import timeout
from ...model.browser.base.driver import BrowserDriver
from ...model.browser.base.settings import BrowserSettings
from ...model.driver_methods import DriverMethods
from .attribute.__main__ import GetAttributeDriverMethods
from .dimensions import get_dimensions
from .element import get_element
from .elements import get_elements
from .elements_by_tag import get_elements_by_tag
from .page_title import get_page_title
from .screen.__main__ import GetScreenSizeDriverMethods
from .screenshot.__main__ import GetScreenshotDriverMethods
from .text import get_text
from .texts import get_texts
from .url.__main__ import GetUrlDriverMethods


class GetDriverMethods(DriverMethods):
    __slots__ = ["attribute", "screen", "screenshot", "url"]

    def __init__(self, browser_driver: BrowserDriver, settings: BrowserSettings) -> None:
        super().__init__(browser_driver, settings)
        self.attribute: GetAttributeDriverMethods = GetAttributeDriverMethods(browser_driver, settings)
        self.screen: GetScreenSizeDriverMethods = GetScreenSizeDriverMethods(browser_driver, settings)
        self.screenshot: GetScreenshotDriverMethods = GetScreenshotDriverMethods(browser_driver, settings)
        self.url: GetUrlDriverMethods = GetUrlDriverMethods(browser_driver, settings)

    def dimensions(self, xpath: str, timeout: int = timeout.DEFAULT) -> tuple[int, int]:
        """Get width and height of element in pixels. Usage:

        width, height = browser.get.dimensions("/element/xpath")"""

        return get_dimensions(self._driver, xpath, timeout)

    def element(self, xpath: str, timeout: int = timeout.DEFAULT) -> object:
        """Get single web element by XPath."""

        return get_element(self._driver, xpath, timeout)

    def elements(self, xpath: str, timeout: int = timeout.DEFAULT) -> list[object]:
        """Get multiple web elements by XPath."""

        return get_elements(self._driver, xpath, timeout)

    def elements_by_tag(self, tag: str, timeout: int = timeout.DEFAULT) -> list[object]:
        """"Get multiple web elements by HTML tag. Examples: "img" as tag for all <img> images, "a" for all <a> links, etc."""

        return get_elements_by_tag(self._driver, tag, timeout)

    def page_title(self) -> str:
        """Get page title of the current page."""

        return get_page_title(self._driver)

    def text(self, xpath: str, timeout: int = timeout.DEFAULT) -> str:
        """Get text from element.

        This method assumes that the text field shouldn't be empty and therefore will retry to get the text (for better support of single-page apps with extended loading time)."""

        return get_text(self._driver, xpath, timeout)

    def texts(self, xpath: str, timeout: int = timeout.DEFAULT) -> list[str]:
        """Get array of texts from elements.

        Assumes that the XPath targets multiple elements."""

        return get_texts(self._driver, xpath, timeout)
